@@ -1,7 +1,7 @@
 from ast import Call
 import asyncio
 import serial
-import socket
+import time
 from enum import IntEnum
 from typing import Any, Callable, List, Tuple
 
@@ -18,6 +18,11 @@ class NotConnectedError(Exception):
 class ConnectionInvalidError(Exception):
     pass
 
+class RequestTimeoutError(Exception):
+    pass
+
+class ShutdownError(Exception):
+    pass
 class DeviceType(IntEnum):
     FP10 = 0
     FP6 = 1
@@ -127,8 +132,6 @@ SENSOR_LABELS = [
 ]
 
 DecoderCB = Callable[[HanazederPacket], Any]
-
-ParseCB = Callable[[Any], Any]
 class ConfigEntry:
     key = 0
 
@@ -139,12 +142,21 @@ class ConfigEntry:
         self.min_value = data[2]
         self.change_step = data[3]
 
+class HanazederRequestState(IntEnum):
+    UNSENT = 0
+    SENT = 1
+    SUCCESSFUL = 2
+    TIMEOUT = 3
+    SHUTDOWN = 4
 class HanazederRequest:
-    def __init__(self, msg_no: int, type: int, cb: ParseCB, decoder: DecoderCB):
+    result = None
+    def __init__(self, msg_no: int, type: int, decoder: DecoderCB):
         self.msg_no = msg_no
-        self.cb = cb
         self.type = type
         self.decoder = decoder
+        self.created = time.monotonic()
+        self.event = asyncio.Event()
+        self.state = HanazederRequestState.UNSENT
 
 class FPProtocol(asyncio.Protocol):
     device = None
@@ -154,15 +166,14 @@ class FPProtocol(asyncio.Protocol):
     def data_received(self, data):
         if self.device.debug:
             print(f'Data received: {data}')
-        for byte in data:
-            packet = self.device.reader.read(byte)
-            if packet:
-                self.device.handle_packet(packet)
+        asyncio.get_event_loop().create_task(self.device.read_bytes(data))
 
     def connection_lost(self, exc):
         print('The server closed the connection')
-        print('Stop the event loop')
-        self.device.loop.stop()
+        self.device.connected = False
+        # Awake all listeners
+        self.device.queue_empty_event.clear()
+
 class HanazederFP:
     HEADER = b'\xEE'
     last_msg_num = 0
@@ -170,15 +181,14 @@ class HanazederFP:
     debug = False
     connection: SerialOrNetwork
     queue: List[HanazederRequest] = []
+    running = True
 
-    def __init__(self, debug=False):
+    def __init__(self, debug=False, request_timeout=2):
         self.debug = debug
         self.loop = asyncio.get_running_loop()
-        self.queue_empty_event = asyncio.Event()
-        # self.serial_port = serial_port
-        # self.address = address
-        # self.port = port
-        # self.timeout = timeout
+        self.queue_lock = asyncio.Lock()
+        self.msg_no_lock = asyncio.Lock()
+        self.request_timeout = request_timeout
     
     async def open(self,
             serial_port="/dev/ttyUSB0",
@@ -197,55 +207,102 @@ class HanazederFP:
         elif address and port:
             (self.connection, proto) = await self.loop.create_connection(FPProtocol, address, port)
             proto.device = self
-            # self.connection = socket.create_connection((address, port), timeout).makefile('rwb')
         else:
             raise ConnectionInvalidError("Specify either address and port or serial port")
         self.reader = HanazederReader(self.connection, self.HEADER, self.debug)
         if serial_port:
             self.loop.add_reader(self.connection, self.read_byte)
+        # Check queue for stuck messages periodically
+        self.loop.create_task(self.check_queue())
     
-    async def wait_for_empty_queue(self):
-        await self.queue_empty_event.wait()
     
-    async def send_msg(self, msg: bytes, cb: ParseCB, decoder: DecoderCB) -> bool:
+    async def get_next_msg_no(self) -> int:
+        async with self.msg_no_lock:
+            msg_no = self.last_msg_num
+            self.last_msg_num = (self.last_msg_num + 1) % 256
+            # Avoid escape byte
+            if self.last_msg_num == 238:
+                self.last_msg_num = 239
+            return msg_no
+    
+    async def send_msg(self, msg: bytes, decoder: DecoderCB) -> HanazederRequest:
         if self.debug:
             print(f'Sending msg {byte_to_hex(msg)}')
-        self.queue.append(HanazederRequest(msg[1], msg[2], cb, decoder))
-        self.queue_empty_event.clear()
+        request = HanazederRequest(msg[1], msg[2], decoder)
+        async with self.queue_lock:
+            self.queue.append(request)
         self.connection.write(msg)
-        # await self.connection.flush()
-        msg_num = self.last_msg_num
-        self.last_msg_num = (self.last_msg_num + 1) % 256
-        return msg_num
+        request.state = HanazederRequestState.SENT
+        return request
+    
+    async def read_bytes(self, bytes):
+        for byte in bytes:
+            packet = self.reader.read(byte)
+            if packet:
+                await self.handle_packet(packet)
     
     def read_byte(self):
+        self.loop.create_task(self.read_byte_async())
+    
+    async def read_byte_async(self):
         byte = self.connection.read(1)
         packet = self.reader.read(byte[0])
         if packet:
-            self.handle_packet(packet)
+            await self.handle_packet(packet)
+    
+    async def check_queue(self):
+        while self.connected:
+            now = time.monotonic()
+            async with self.queue_lock:
+                for index, request in enumerate(self.queue):
+                    if now - request.created > self.request_timeout:
+                        print(f'Request #{request.msg_no} has timed out')
+                        # TODO: resend
+                        request.state = HanazederRequestState.TIMEOUT
+                        request.event.set()
+                        self.queue.pop(index)
+            await asyncio.sleep(self.request_timeout)
 
-    def handle_packet(self, packet: HanazederPacket):
+    def shutdown(self):
+        self.running = False       
+        for request in self.queue:
+            request.state = HanazederRequestState.SHUTDOWN
+            request.event.set()
+
+
+    async def handle_packet(self, packet: HanazederPacket):
         if self.debug:
-            print(f'Packet read: {packet}')
+            packet_debug = ""
+            for req in self.queue:
+                packet_debug = f"{packet_debug} #{req.msg_no}"
+            print(f'Packet read #{packet.msg_no} type {packet.msg_type}: {packet.msg}.')
+            print(f'Queue: {packet_debug}')
         found = False
-        for index, req in enumerate(self.queue):
-            if req.msg_no == packet.msg_no:
-                req.cb(req.decoder(packet))
-                found = True
-                self.queue.pop(index)
-                break
-        if not found:
-            print(f"Couldn't find message {packet.msg_no} in queue!")
-        else:
-            if len(self.queue) == 0:
-                self.queue_empty_event.set()
+        async with self.queue_lock:
+            for index, req in enumerate(self.queue):
+                if req.msg_no == packet.msg_no:
+                    req.result = req.decoder(packet)
+                    found = True
+                    req.state = HanazederRequestState.SUCCESSFUL
+                    req.event.set()
+                    self.queue.pop(index)
+                    break
+            if not found:
+                print(f"Couldn't find message {packet.msg_no} in queue!")
 
+    async def handle_req_response(self, req: HanazederRequest):
+        if req.state == HanazederRequestState.TIMEOUT:
+            raise RequestTimeoutError()
+        elif req.state == HanazederRequestState.SHUTDOWN:
+            raise ShutdownError()
     
-    def create_read_information_msg(self) -> bool:
-        return hanazeder_encode_msg(self.HEADER, self.last_msg_num, b'\x01\x00')
+    async def create_read_information_msg(self) -> bool:
+        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), b'\x01\x00')
     
-    async def read_information(self, cb: Callable[[None], Any]):
-        await self.send_msg(self.create_read_information_msg(), cb, self.parse_information_packet)
+    async def read_information(self):
+        req = await self.send_msg(await self.create_read_information_msg(), self.parse_information_packet)
+        await req.event.wait()
+        await self.handle_req_response(req)
     
     def parse_information_packet(self, msg: HanazederPacket):
         response = msg.msg
@@ -256,24 +313,28 @@ class HanazederFP:
         if (len(response) >= 5):
             self.version = f'{response[3]}.{response[4]}'
     
-    def create_read_sensor_msg(self, idx: int) -> bytes:
-        request = bytes(b'\x04\x01') + idx.to_bytes(1, byteorder='little')
-        return hanazeder_encode_msg(self.HEADER, self.last_msg_num, request)
+    async def create_read_sensor_msg(self, idx: int) -> bytes:
+        request_bytes = bytes(b'\x04\x01') + idx.to_bytes(1, byteorder='little')
+        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), request_bytes)
     
-    async def read_sensor(self, idx: int, cb: Callable[[float], Any]):
+    async def read_sensor(self, idx: int) -> float:
         if not self.connected:
             raise NotConnectedError()
-        await self.send_msg(self.create_read_sensor_msg(idx), cb, self.parse_sensor_packet)
+        
+        req = await self.send_msg(await self.create_read_sensor_msg(idx), self.parse_sensor_packet)
+        await req.event.wait()
+        await self.handle_req_response(req)
+        return req.result
     
     def parse_sensor_packet(self, msg: HanazederPacket) -> float:
         value = hanazeder_decode_num(self.HEADER, msg.msg)
         return value
     
-    def create_read_config_block_msg(self, start: int, count: int) -> bytes:
+    async def create_read_config_block_msg(self, start: int, count: int) -> bytes:
         request = bytes(b'\x07\x03') + start.to_bytes(2, byteorder='little') + count.to_bytes(1, byteorder='little')
-        return hanazeder_encode_msg(self.HEADER, self.last_msg_num, request)
+        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), request)
     
-    async def read_config_block(self, start: int, count: int, cb: Callable[[List[ConfigEntry]], Any]):
+    async def read_config_block(self, start: int, count: int) -> List[ConfigEntry]:
         if not self.connected:
             raise NotConnectedError()
         
@@ -285,32 +346,42 @@ class HanazederFP:
                 entry = ConfigEntry(start + x, chunk)
                 entries.append(entry)
             return entries
-        await self.send_msg(self.create_read_config_block_msg(start, count), cb, parse_config_block_packet)
+        req = await self.send_msg(await self.create_read_config_block_msg(start, count), parse_config_block_packet)
+        await req.event.wait()
+        await self.handle_req_response(req)
+        return req.result
 
-    def create_read_sensor_name_msg(self, idx: int) -> bytes:
+    async def create_read_sensor_name_msg(self, idx: int) -> bytes:
         request = bytes(b'\x13\x01') + idx.to_bytes(1, byteorder='little')
-        return hanazeder_encode_msg(self.HEADER, self.last_msg_num, request)
+        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), request)
     
-    async def read_sensor_name(self, idx: int, cb: Callable[[str], Any]):
+    async def read_sensor_name(self, idx: int) -> str:
         if not self.connected:
             raise NotConnectedError()
-        await self.send_msg(self.create_read_sensor_name_msg(idx), cb, self.parse_sensor_name_packet)
+
+        req = await self.send_msg(await self.create_read_sensor_name_msg(idx), self.parse_sensor_name_packet)
+        await req.event.wait()
+        await self.handle_req_response(req)
+        return req.result
     
     def parse_sensor_name_packet (self, msg: HanazederPacket) -> str:
         if msg.msg and len(msg.msg) > 1:
             return msg.msg[1:].decode('ascii', errors='ignore').strip()
     
-    def create_read_debug_block_msg(self, start: int, count: int) -> bytes:
+    async def create_read_debug_block_msg(self, start: int, count: int) -> bytes:
         request = bytes(b'\x20\x03') + dec_to_bytes(start) + count.to_bytes(1, byteorder='little')
-        return hanazeder_encode_msg(self.HEADER, self.last_msg_num, request)
+        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), request)
     
-    async def read_debug_block(self, start: int, count: int, cb: Callable[[Any], Any], decoder: DecoderCB) -> bytes:
+    async def read_debug_block(self, start: int, count: int, decoder: DecoderCB) -> Any:
         if not self.connected:
             raise NotConnectedError()
-        await self.send_msg(self.create_read_debug_block_msg(start, count), cb, decoder)
+        req = await self.send_msg(await self.create_read_debug_block_msg(start, count), decoder)
+        await req.event.wait()
+        await self.handle_req_response(req)
+        return req.result
     
-    async def read_energy(self, cb: Callable[[Tuple[int, int, int]], Any]) -> EnergyReading:
-        await self.read_debug_block(313, 8, cb, self.parse_energy_packet)
+    async def read_energy(self) -> EnergyReading:
+        return await self.read_debug_block(313, 8, self.parse_energy_packet)
     
     def parse_energy_packet(self, msg: HanazederPacket) -> Tuple[int, int, int]:
         total = hanazeder_decode_num(self.HEADER, msg.msg[0:2])
